#ifndef __TERM_CPP
#define __TERM_CPP

#include <string>

namespace terminal {

/**
 * @brief represents state of input form
 */
struct state {
    std::string keys;
    unsigned int cursorPos = 0;
};

/**
 * @brief removes current character from state.
 * 
 * @param state
 * @return true if character was removed
 */
bool key_backspace(state& state) {
    if (state.cursorPos <= 0) return false;
    state.keys.erase(state.cursorPos - 1, 1);
    state.cursorPos--;
    return true;
}

/**
 * @brief moves cursor one characer to the left
 * 
 * @param state
 * @return true if moved
 */
bool key_left(state& state) {
    if (state.cursorPos <= 0) return false;
    state.cursorPos--;
    return true;
}

/**
 * @brief moves cursor one character to the right
 * 
 * @param state
 * @return true if moved
 */
bool key_right(state& state) {
    if (state.cursorPos >= state.keys.length()) return false;
    state.cursorPos++;
    return true;
}

/**
 * @brief inserts character to current state
 * 
 * @param state form state
 * @param key character to insert
 */
void key_insert(state& state, unsigned char key) {
    state.keys.insert(state.cursorPos, std::string(1, key));
    state.cursorPos++;
}

} //namespace terminal

#endif